/* $Id: e2p_cpbar.c 1509 2006-01-23 11:45:42Z tpgww $
 * **************************************
 *
 * Emelcopybar - a plugin for emelFM2
 * by Florian Zaehringer
 * with help from tooar and tom to port it to Gtk+2
 *
 * source of source-code and inspiration: 
 * emelFM by Michael Clark
 *
 * **************************************
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 ***************************************/
#include "emelfm2.h"
#include <string.h>
#include <pthread.h>
#include "e2_plugins.h"
#include "e2_dialog.h"
#include "e2_option.h"
#include "e2_filelist.h"
#include "e2_task.h"

//max. no. of _bytes_, at the end of a copied item's source-path
//and dest-path, shown in the progress dialog
//paths longer than this will be 'ellipsized'
#define MAX_CHAR 55
//size of buffer for progress dialog
//make this >= (MAX_CHAR+4)*2 + space for rest of message text
#define MAX_MSG 220
//largest interval (usec) between progress window updates
#define MAX_UPDATE_INTERVAL 250000

typedef struct
{
	guint64 count;	//count of items to be processed, in active pane
	guint64 totalsize;	//aggregate apparent size of items to be processed
} E2_BarData;

typedef struct
{
	pthread_mutex_t mutex;  //protects access to these data
	pthread_cond_t cond;   //signals change
	gchar *dlocal;	//localised destination (maybe temp) path of item copied
	guint64 done_size;
	guint64 refresh_interval;
} E2_ProgressData;

typedef struct
{
//	pthread_mutex_t mutex;  //protects access to these data
	gchar *slocal;	//localised source path of item copied
	gchar *dlocal;	//localised destination path of item copied
	gboolean completed;	//set TRUE when the backend task is completed (valid or not)
	gboolean result;	//value returned by backend task function
} E2_ActionData;

typedef struct
{
	GtkWidget *window;
	GtkWidget *label;
	GtkWidget *label2;
	GtkWidget *progbar;
} E2_BarWindowData;

/**
 * @brief update items' count and total size
 * This is a callback for a treewalk function
 * @param local_name path of item reported by the walker, localised string
 * @param statbuf pointer to struct stat with data about @a local_name
 * @param status code from the walker, indicating what type of item is being reported
 * @param twdata pointer to tw data struct
 * @return E2TW_CONTINUE always
 **/
static E2_TwResult
_e2p_cpbar_twcb (const gchar *local_name,
	const struct stat *statbuf, E2_TwStatus status, E2_BarData *twdata)
{
	gchar *utf;
	gchar *msg;
	switch (status)
	{
		case E2TW_DL:	//directory, not opened due to tree-depth limit
		case E2TW_DM:	//directory, not opened due to different file system
		case E2TW_DNR:	//unreadable directory
			utf = F_FILENAME_FROM_LOCALE ((gchar *)local_name);
			msg = g_strdup_printf ("%s %s\n",
				_("Warning: couldn't open directory"), utf);
			e2_output_print_default (msg);
			e2_output_print_errno ();
			F_FREE (utf);
			g_free (msg);
		case E2TW_D:	//directory
		case E2TW_F:	//not directory or link
		case E2TW_SL:	//symbolic link
		case E2TW_SLN:	//symbolic link naming non-existing file
			twdata->count++;
			twdata->totalsize += statbuf->st_size;
			break;
		case E2TW_NS:	//un-statable item
			//WARN ??
			twdata->count++;
//		case E2TW_DP:	//directory, finished
		default:
			break;
	}
	return E2TW_CONTINUE;
}
/**
 * @brief 'shorten' string @a string if it is longer than allowed
 * @a dots is set to "..." if @a string is too long, or else ""
 *
 * @param string utf string which may need to be shortened
 * @param dots pointer to store for utf string which will preceed @a string, when it is displayed
 * @param string_diff pointer to store for the no. of bytes to be omitted from the start of @a string, when it is displayed
 *
 * @return
 **/
static void
_e2p_cpbar_shorten (gchar *string, gchar **dots, gint *string_diff)
{
	if (strlen (string) > MAX_CHAR)
	{
		//nominally we will drop this many bytes off the start of string
		gint trunc = strlen (string) - MAX_CHAR;
		//but make sure we don't cut off in the middle of a char
		gchar *c = g_utf8_find_next_char (string+trunc-1, NULL);
		*string_diff = c - string;
		*dots = "...";
	}
	else
	{
		*string_diff = 0;
		*dots = "";
	}
	return;
}
/**
 * @brief determine how much progress has been made on the current item
 *
 * This is the thread function which provides progressive updates
 * during the course of copying an item.
 * The thread continues until it is cancelled by the main thread
 *
 * @param data pointer to struct with data which is needed here
 *
 * @return NULL
 **/
static void *
_e2p_cpbar_progress (E2_ProgressData *data)
{
	gchar *local;
	struct stat sb;
//	printd (DEBUG, "monitor thread underway");
	pthread_testcancel ();	//CHECKME needed ?

	while (TRUE)
	{
		if (pthread_mutex_lock (&(data->mutex)))
		{
			printd (WARN, "Lock error!");
		}
		//get a local copy of the present item's destination path
		//(that will be NULL if the action thread has finished already)
		local = (data->dlocal == NULL) ? NULL : g_strdup (data->dlocal);

		if (pthread_mutex_unlock (&(data->mutex)))
		{
			printd (WARN,"Unlock error!");
		}

		if (local != NULL)
		{	//we're still underway
			//determine how much of current item has been copied
			E2_BarData pdata = { 0,0 };
			if (!e2_fs_lstat (local, &sb))
			{
				if (S_ISDIR (sb.st_mode))
					e2_fs_tw (local, _e2p_cpbar_twcb, &pdata, -1, E2TW_PHYS);
				else
					pdata.totalsize = sb.st_size;
			}
			g_free (local);

			if (pthread_mutex_lock (&(data->mutex)))
			{
				printd (WARN, "Lock error!\n");
			}
			//update size data for the main thread
			data->done_size = pdata.totalsize;
//			printd (DEBUG, "progress is %li", pdata.totalsize);

			if (pthread_cond_signal (&data->cond))
			{
				printd (WARN,"Signal error!");
			}

			if (pthread_mutex_unlock (&(data->mutex)))
			{
				printd (WARN,"Unlock error!");
			}
		}
		//CHECKME is this a cancellation point ?
		g_usleep (data->refresh_interval);	//delay between updates
	}
	return NULL;
}
/**
 * @brief perform the copy
 *
 * This is the thread function which performs the copy task.
 * The thread simply exits when completed
 *
 * @param data pointer to struct with data which is needed here
 *
 * @return NULL
 **/
static void *
_e2p_cpbar_action (E2_ActionData *data)
{
//	printd (DEBUG, "action thread underway");
	data->result = e2_task_backend_copy	//no background-copy option in this context
		(data->slocal, data->dlocal);
	data->completed = TRUE;
	return NULL;
}
/**
 * @brief update progress window details, and copy item @a src to @a dest
 *
 * @param src path of item to be copied, utf-8 string
 * @param dest new path of copied item, utf-8 string
 * @param dest_dir the directory part of @a dest, utf-8 string
 * @param breakflag pointer to store for abort-flag
 * @param bdata pointer to bar data struct holding progress data
 * @param tdata pointer to bar data struct holding totals data
 * @param wdata pointer to info-window data struct
 *
 * @return TRUE if the copy operation completed successfully
 **/
static gboolean _e2p_cpbar_exec (gchar *src, gchar *dest, gchar *dest_dir,
	gboolean *breakflag,
	E2_BarData *bdata, E2_BarData *tdata, E2_BarWindowData *wdata)
{
	gchar labeltext[MAX_MSG];	//utf-8 string, top line of progress dialog
	gchar progresstext[64];	//utf-8 string, middle line of progress dialog
	gchar *slocal, *dlocal;	//localised strings
	gchar *src_dots, *dest_dots;	//set to "...", or "" if name not shortened
	gint src_diff, dest_diff;	//offsets into src, dest when they are shortened
	gfloat fraction;

	_e2p_cpbar_shorten (src, &src_dots, &src_diff);
	_e2p_cpbar_shorten (dest_dir, &dest_dots, &dest_diff);	//FIXME do this only once

	g_snprintf (labeltext, sizeof (labeltext),
		_("copying %s%s\nto %s%s"),	//\nthis is item %llu of %llu"),
		src_dots, src+src_diff,
		dest_dots, dest_dir+dest_diff);
//		bdata->count+1, tdata->count);
	gtk_label_set_text (GTK_LABEL (wdata->label), labeltext);

	gchar *progress_format = _("%.2f MB of total %.2f MB  (%.0f\%%)");
	
	slocal = F_FILENAME_TO_LOCALE (src);
	dlocal = F_FILENAME_TO_LOCALE (dest);

	//get size of item to be copied
	E2_BarData pdata = { 0,0 };
	e2_fs_tw (slocal, _e2p_cpbar_twcb, &pdata, -1, E2TW_PHYS);

    //create action thread
#ifdef LATE_DELETE
	//work with a temp name so that the backend never uses
	//its own (which would prevent destination-monitoring)
	gchar *templocal = e2_utils_get_tempname (dlocal);
	E2_ActionData a_data = { slocal , templocal, FALSE, FALSE };
#else
	E2_ActionData a_data = { slocal , dlocal, FALSE, FALSE };
#endif
	pthread_t action_thread_id;
	gint status = pthread_create (&action_thread_id, NULL,
		(void *) _e2p_cpbar_action, &a_data);
	if (status != 0)
	{
		printd (WARN,"action-thread-create error!");
//		if (errno == EAGAIN)
//			printd (WARN, "not enough system resources");
		return FALSE;
	}
	//wait for task to initiate, at least
	g_usleep (50000);
	//create progress monitor, if task not finished already
	if (!a_data.completed)
	{
		gtk_widget_show (wdata->window);
		//create monitor thread
		E2_ProgressData m_data;
		pthread_mutex_init (&(m_data.mutex), NULL);
		pthread_cond_init (&(m_data.cond), NULL);
#ifdef LATE_DELETE
		//get the temp name now in use for copying the item
		m_data.dlocal = templocal;
#else
		m_data.dlocal = dlocal;
#endif
		m_data.done_size = 0;
		//rough approach to setting the reporting interval (usec)
		m_data.refresh_interval = pdata.totalsize * 10;
		if (m_data.refresh_interval > MAX_UPDATE_INTERVAL)
			m_data.refresh_interval = MAX_UPDATE_INTERVAL;	//with a cap
		pthread_t mon_thread_id;
		status = pthread_create (&mon_thread_id, NULL,
			(void *) _e2p_cpbar_progress, &m_data);
		if (status != 0)
		{
			printd (WARN,"monitor-thread-create error!");
			return FALSE;
		}

		//update progress when the monitor thread reports
		guint64 progress;
		while (!a_data.completed)	//loop until the action thread is completed
		{
			if (pthread_mutex_lock (&(m_data.mutex)))
			{
				printd (WARN, "Lock error!");
			}
			while (m_data.done_size == 0)
			{
				status = pthread_cond_wait (&(m_data.cond), &(m_data.mutex));
			}
			if (status != 0)
			{
				printd (WARN,"Wait error!");
			}

			progress = m_data.done_size + bdata->totalsize;
			m_data.done_size = 0;	//reset

			if (pthread_mutex_unlock (&(m_data.mutex)))
			{
				printd (WARN,"Unlock error!");
			}

			if (*breakflag) //user wants to abort
			{
				if (pthread_cancel (mon_thread_id))
				{
					printd (WARN,"Thread cancel error!\n");
				}
				//cancel the task thread
				if (pthread_cancel (action_thread_id))
				{
					printd (WARN,"Thread cancel error!\n");
				}
				//cleanup anything part-copied
#ifdef LATE_DELETE
				e2_task_backend_delete (templocal);
				g_free (templocal);
#else
				e2_task_backend_delete (dlocal);
#endif
				F_FREE (slocal);
				F_FREE (dlocal);
				return FALSE;
			}

			fraction =  (gdouble) progress / tdata->totalsize;
			g_snprintf (progresstext, sizeof (progresstext), progress_format,
				progress / 1048576.0, tdata->totalsize / 1048576.0, 
				fraction * 100.0);

			gtk_label_set_text (GTK_LABEL (wdata->label2), progresstext);
		//	gtk_progress_bar_set_text (GTK_PROGRESS_BAR (wdata->progbar), progresstext);
			//deal with rounding errors
			if (fraction > 1.0)
				fraction = 1.0;
			gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (wdata->progbar), fraction);
			WAIT_FOR_EVENTS;	//make sure window contents are updated
		}

		pthread_join (action_thread_id, NULL);	//make sure action finishes before we move on
		gtk_widget_hide (wdata->window);

		//immediate signal to monitor thread to stop processing
		if (pthread_mutex_lock (&(m_data.mutex)))
		{
			printd (WARN, "Lock error!");
		}
		m_data.dlocal = NULL;
		if (pthread_mutex_unlock (&(m_data.mutex)))
		{
			printd (WARN,"Unlock error!");
		}
		//cancel the monitor thread
		if (pthread_cancel (mon_thread_id))
		{
			printd (WARN,"Thread cancel error!\n");
		}
	}

	//need to slow down to avoid hangs
	//this approach is empirical, needs to be validated for various systems
	if (bdata->count > 300 && bdata->count%5 == 0)
		g_usleep (50000);	//small delay to allow filesystem to complete its business

	if (a_data.result) //copy succeeded
	{
#ifdef LATE_DELETE
		e2_task_backend_rename (templocal, dlocal);
#endif
		//update the progressive totals
//		bdata->count += pdata.count;
		bdata->count++;
		bdata->totalsize += pdata.totalsize;
		fraction =  (gdouble) bdata->totalsize / tdata->totalsize;
		//and show them
		g_snprintf (progresstext, sizeof (progresstext), progress_format,
			bdata->totalsize / 1048576.0, tdata->totalsize / 1048576.0, 
			fraction * 100.0);

		gtk_label_set_text (GTK_LABEL (wdata->label2), progresstext);
//		gtk_progress_bar_set_text (GTK_PROGRESS_BAR (wdata->progbar), progresstext);
		//deal with rounding errors
		if (fraction > 1.0)
			fraction = 1.0;
		gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (wdata->progbar), fraction);
		WAIT_FOR_EVENTS;	//make sure window contents are updated
	}
	else
	{
		//cleanup anything part-copied
#ifdef LATE_DELETE
		e2_task_backend_delete (templocal);
#else
		e2_task_backend_delete (dlocal);
#endif
	}
#ifdef LATE_DELETE
	g_free (templocal);
#endif
	F_FREE (slocal);
	F_FREE (dlocal);
	return (a_data.result);
}
/**
 * @brief callback for stop-button press and window-close event
 *
 * @param widget the activated widget which triggered the callback
 * @param flag pointer to store for flag which will abort the process when set to TRUE;
 *
 * @return TRUE always
 **/
static gboolean
_e2p_cpbar_break_cb (GtkWidget *widget, gboolean *flag)
{
	*flag = TRUE;
	return TRUE;
}
/**
 * @brief setup progress display window, and process each selected item
 *
 * @return
 **/
static void 
_e2p_cpbar (void)
{
	GtkWidget *separator;
	GtkWidget *vbox, *hbox;
	GList *snapshot_list, *snapshot_base;
	gchar *src_dir, *dest_dir, *src, *dest;	//pointers to utf strings
	gchar *slocal, *dlocal;	//pointers to localised strings

	FileInfo *info = e2_fileview_get_selected_first_local (curr_view);
	if (info == NULL)
		return;

	dlocal = F_FILENAME_TO_LOCALE (other_view->dir);
	if (e2_fs_access (dlocal, W_OK))
	{
		e2_output_print_default (_("\nI recently spoke to your filesystem and it told me you\n"
			"aren't allowed to write there. You should check that...\n"));
		F_FREE (dlocal);
		return;
	}
	F_FREE (dlocal);

	//setup the information window
	gboolean breakflag = FALSE;	//this is for aborting the process
	E2_BarWindowData windowdata;
	windowdata.window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	E2_DIALOG_TITLE (GTK_WINDOW (windowdata.window), _("copying"));
	gtk_window_set_position (GTK_WINDOW (windowdata.window), GTK_WIN_POS_CENTER);
	g_signal_connect (G_OBJECT (windowdata.window), "delete_event",
		G_CALLBACK (_e2p_cpbar_break_cb), &breakflag);

	vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), E2_PADDING_LARGE);
	gtk_container_add (GTK_CONTAINER (windowdata.window), vbox);
	gtk_widget_show (vbox);

	hbox = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
//	windowdata->label = gtk_label_new (_("Initializing emelcopybar...")); no point, not visible for long enough
	windowdata.label = gtk_label_new ("");
	gtk_box_pack_start (GTK_BOX (hbox), windowdata.label, FALSE, FALSE, 0);
//	gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_CENTER);
	gtk_widget_show (windowdata.label);
	gtk_widget_show (hbox);

	separator = gtk_hseparator_new ();
	gtk_box_pack_start (GTK_BOX (vbox), separator, TRUE, TRUE, 0);
	gtk_widget_show (separator);

	windowdata.label2 = gtk_label_new ("");
	gtk_box_pack_start (GTK_BOX (vbox), windowdata.label2, TRUE, TRUE, 0);
	gtk_label_set_justify (GTK_LABEL (windowdata.label2), GTK_JUSTIFY_CENTER);
	gtk_widget_show (windowdata.label2);

	hbox = gtk_hbox_new (FALSE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (hbox), E2_PADDING_LARGE);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
	windowdata.progbar = gtk_progress_bar_new ();
//	gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (progbar), 0);
	gtk_box_pack_start (GTK_BOX (hbox), windowdata.progbar, TRUE, TRUE, 0);
	gtk_widget_show_all (hbox);

	hbox = gtk_hbox_new (FALSE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (hbox), E2_PADDING_SMALL);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
	e2_button_add_end (hbox, FALSE, 0, _("_Stop"), GTK_STOCK_STOP,
		_("Abort the copying"), _e2p_cpbar_break_cb, &breakflag);
	gtk_widget_show_all (hbox);

	//accumulate count and total size of src item(s)
	//(the recursed count is not used)
	E2_BarData totaldata = { 0, 0 };
	src_dir = F_FILENAME_TO_LOCALE (curr_view->dir);
	gint topcount = 0;
	//prevent loss of info items' data while processing
	e2_filelist_disable_refresh ();
	snapshot_list = snapshot_base = e2_fileview_get_selected_local (curr_view);
	for (; snapshot_list != NULL ; snapshot_list = snapshot_list->next)
	{
		slocal = e2_utils_strcat (src_dir, ((FileInfo *)snapshot_list->data)->filename);
		e2_fs_tw (slocal, _e2p_cpbar_twcb, &totaldata, -1, E2TW_PHYS);
		g_free (slocal); //glib used on localised strings, should be ok
		topcount++;
	}
	e2_filelist_enable_refresh ();
	F_FREE (src_dir);
	g_list_free (snapshot_base);

	//get a copy of dirs, in case user changes dir
	src_dir = g_strdup (app.pane_current->path);
	dest_dir = g_strdup (app.pane_other->path);

	gboolean check = e2_option_bool_get ("confirm-overwrite");
	E2_BarData progressdata = { 0, 0 };

	e2_widget_set_cursor (GDK_WATCH);
	e2_filelist_disable_refresh ();

	GPtrArray *names = e2_fileview_get_selected (curr_view);
	guint count;
	E2_SelectedItemInfo **iterator = (E2_SelectedItemInfo **) names->pdata;
//	const_base = const_list;
	OW_ButtonFlags extras = (totaldata.count > 1) ? BOTHALL : NONE;
	totaldata.count = topcount;	//report only the count of selected items

	for (count=0; count < names->len; count++, iterator++)
	{
		if (breakflag)
			break;	//user pressed stop btn or closed info window
		//src_dir, dest_dir have trailing "/"
		src = g_strconcat (src_dir, (*iterator)->filename, NULL);
		dest = g_strconcat (dest_dir, (*iterator)->filename, NULL);
		dlocal = F_FILENAME_TO_LOCALE (dest);

		if (check && e2_fs_access2 (dlocal) == 0)
		{
			e2_filelist_enable_refresh ();  //allow updates while we wait
			DialogButtons result = e2_dialog_ow_check (dest, extras);
			e2_filelist_disable_refresh ();
			switch (result)
			{
				case YES_TO_ALL:
					check = FALSE;
				case OK:
					_e2p_cpbar_exec (src, dest, dest_dir, &breakflag, &progressdata, &totaldata, &windowdata);
					break;
				case CANCEL:
					break;
				default:
					result = NO_TO_ALL;
					break;
			}
			if (result == NO_TO_ALL)
			{
				g_free (src);
				g_free (dest);
				F_FREE (dlocal);	//finished with dlocal for this item
				break;
			}
		}
		else  //no overwrite, or don't care
		{
			_e2p_cpbar_exec (src, dest, dest_dir, &breakflag, &progressdata, &totaldata, &windowdata);
		}
		g_free (src);
		g_free (dest);
		F_FREE (dlocal);	//finished with dlocal for this item
	}
	gtk_widget_destroy (windowdata.window);
	e2_fileview_clean_selected (names);
	g_free (src_dir);
	g_free (dest_dir);

	//CHECKME E2_INCLIST - do other view, & current if it's same
	e2_filelist_check_dirty (GINT_TO_POINTER(1));

	e2_widget_set_cursor (GDK_LEFT_PTR);
		e2_filelist_enable_refresh ();
}

//aname must be confined to this module
static gchar *aname;
/**
 * @brief initialize this plugin
 *
 * @param p pointer to plugin data struct to be populated
 *
 * @return TRUE if the intialization succeeded
 **/
gboolean
init_plugin (Plugin *p)
{
#define ANAME "cpbar"
	aname = _("cpbar");

	p->signature = ANAME VERSION;
	p->menu_name = _("_Copy");
	p->description = _("Copy selected item(s), with displayed progress details");
	p->icon = "plugin_copy_"E2IP".png";  //use icon file pathname if appropriate

	if (p->plugin_cb == NULL)
	{
		p->plugin_cb = _e2p_cpbar;  //set plugins menu item function
//		E2_Action* (*action_register) () = (*p->api_lookup) (E2API_ACTION_REGISTER);
		//no need to free this
		gchar *action_name = g_strconcat (PLUGIN,".",aname,NULL);
		//NULL = no data, FALSE = no argument for the function
//		(*action_register)
		e2_action_register_simple (action_name, E2_ACTION_TYPE_ITEM, _e2p_cpbar, NULL, FALSE);
		return TRUE;
	}
	return FALSE;
}
/**
 * @brief unload this plugin
 *
 * @param p pointer to data struct for the plugin
 *
 * @return
 **/
void
unload_plugin (Plugin *p)
{
//	if (p->api_lookup != NULL)
//	{
//		gboolean (*action_unregister) () = (*p->api_lookup) (E2API_ACTION_UNREGISTER);
		gchar *action_name = g_strconcat (PLUGIN,".",aname,NULL);
//		(*action_unregister)
		e2_action_unregister (action_name);
		g_free (action_name);
//	}
}
